#
# MIT License
#
# Copyright (c) 2017-2019 Paul Taylor
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import subprocess
import os

from gpg_vault import errors, log, config, file, utils

def decryptFile(vpath, destPath, pp64):
    log.verbose(f"decyptFile vpath={vpath} destPath={destPath}")
    passphrase = utils.b642str(pp64)
    log.sensitive(f"passphrase |{passphrase}|")
    args = ['gpg', '--quiet', '--batch', '--ignore-mdc-error', '--decrypt', '--passphrase-fd', '0']
    if destPath is not None:
        args.extend(['--output', destPath])
    args.append(vpath)
    log.verbose(f"args: {args}")
    gpgErrorLog = os.path.join(utils.getVaultDir(), "gpg.error.log")
    try:
        gpgError = open(gpgErrorLog, "a")
    except IOError:
        raise errors.VaultError(f"Unable to open gpg error log file '{gpgErrorlog}'")
    try:
        gpg = subprocess.Popen(args=args, shell=False, stdin=subprocess.PIPE, stderr=gpgError)
        gpg.stdin.write(passphrase.encode('utf-8'))
        gpg.stdin.close()
        gpg.wait()
        if gpg.returncode != 0:
            raise errors.VaultSecurityError("Failed to decrypt file '" + vpath + "'")
    finally:
        log.verbose("closing gpg error file")
        if gpgError:
            gpgError.close


def encryptFile(srcPath, destPath, pp64):
    log.verbose(f"encyptFile srcPath={srcPath} destPath={destPath}")
    passphrase = utils.b642str(pp64)
    log.sensitive(f"passphrase |{passphrase}|")
    if os.path.exists(destPath):
        file.deletePath(destPath)
    args = ['gpg', '--quiet', '--batch',
            '-c', '--symmetric', '--passphrase-fd', '0',
            '--output', destPath, srcPath]
    log.verbose(f"args: {args}")
    gpgErrorLog = os.path.join(utils.getVaultDir(), "gpg.stderr.log")
    log.info(f"encrypting file {srcPath} to {destPath}")
    try:
        gpgError = open(gpgErrorLog, "a")
    except IOError:
        raise errors.VaultError("Unable to open gpg error log file '" + gpgErrorLog + "'")
    try:
        gpg = subprocess.Popen(args=args, shell=False, stdin=subprocess.PIPE, stderr=gpgError)
        gpg.stdin.write(passphrase.encode('utf-8'))
        gpg.stdin.close()
        gpg.wait()
        if gpg.returncode != 0:
            raise errors.VaultSecurityError("Failed to encrypt file '" + srcPath + "'", [])
    finally:
        log.verbose("closing gpg error file")
        if gpgError:
            gpgError.close



#
# MIT License
#
# Copyright (c) 2017-2019 Paul Taylor
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


import os
import sys
import getopt
import configparser

from .  import utils, log, __version__

WORK_DIR_NAME = ".gpg_vault"

DEFAULT_CONFIG = {

        'general': {
            'default_group': 'default',
            'verbose': False,
            'quiet': False,
            'sensitive': False,
            'kill_server': False,
            'sensitive': False
        },

        'server': {
            'port': 61270,
            'timeout_reset_all': 60 * 30,
            'timeout_reset_one': 60 * 10
        },

        'commands': {
            'cat': 'cat {}',
            'edit': 'vi {}',
            'open': 'open {}',
            'delete_file': 'rm -f {}',
            'delete_dir': 'rm -rf {}',
            'clear': None,
            'encrypt': None
        },

        'internal': {
            'vexts': ['.gpg', '.pgp', '.v'],
            'ext.default': '.txt',
            'vext.default': '.gpg',
            'ext.backup': '.bak',
            'tmp.dir.prefix': 'gpg_vault_'
        }
}


global CONFIG
CONFIG = {
    'verbose': False
}


def dump_config():
    log.verbose("configuration:")
    log.verbose(f"\n{utils.dict2str(CONFIG)}")


def init(argv):

    global CONFIG
    CONFIG = DEFAULT_CONFIG

    vdir = utils.getVaultDir()
    config_file = "%s/%s" % (vdir, "config")

    config = configparser.SafeConfigParser()

    if os.path.exists(config_file):
        try:
            config.read(config_file)

        except Exception as e:
            log.error(f"cannot load config file, exception {e}")
            exit(1)

    for k in CONFIG:
        if isinstance(CONFIG[k], dict):
            CONFIG[k] = utils.merge(CONFIG[k], getSection(config, k))

    process_args(argv)

    dump_config()


def process_args(argv):

    global CONFIG
    try:
        opts, args = getopt.getopt(argv, "g:hvqk", ["group=", "help", "verbose", "quiet", "kill", "version"])
    except getopt.GetoptError:
        usage()

    CONFIG['general']['group'] = CONFIG['general']['default_group']

    for opt, arg in opts:
        if opt in ("-g", "--group"):
            CONFIG['general']['group'] = arg
        elif opt in ("-h", "--help"):
            usage()
        elif opt in ("-v", "--verbose"):
            CONFIG['general']['verbose'] = True
        elif opt in ("-k", "--kill"):
            CONFIG['general']['kill_server'] = True
        elif opt in ("-q", "--quiet"):
            CONFIG['general']['quiet'] = True
        elif opt in ("--version"):
            print(f"{__version__}")

    CONFIG['files'] = args


def getSection(config, name):
    m = {}
    try:
        section = config.items(name)
        for v in section:
            m[v[0]] = v[1]
    except configparser.NoSectionError as e:
        log.verbose(f"no config section {name}")
    return m


def usage():

    cmd = os.path.basename(sys.argv[0])
    print(f"""

usage: {cmd} [options] [<path>]

  -g --group     Credentail group to use (default is 'default').

  -l --kill      Kill server after running.

  -v --verbose   Verbose logging.

  -q --quiet     Reduced logging.

  -h --help      Display this help.

  --version      Display version information.

For detailed information see https://github.com/ptaylor/gpg-vault/blob/v{__version__}/README.md

""")
    sys.exit(1)



#
# MIT License
#
# Copyright (c) 2017-2019 Paul Taylor
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os
import getpass

from gpg_vault import config, utils, log, client, crypto, errors, file

def validateVPaths(path):
    (plainPath, vpath) = validateVPath(path)
    log.verbose(f"validating plainPath={plainPath}, vpath={vpath}")
    if os.path.islink(plainPath):
        raise errors.VaultError("'" + plainPath + "' is a symbolic link")
    if os.path.exists(plainPath) and os.path.exists(vpath):
        if os.path.getmtime(plainPath) > os.path.getmtime(vpath):
            older = f"{plainPath} is more recent"
        else:
            older = f"{vpath} is more recent"
        raise errors.VaultError(f"both '{plainPath}' and '{vpath}' exist ({older})")
    if os.path.islink(vpath):
        raise errors.VaultError(f"'{vpath}' is a symbolic link")
    return (plainPath, vpath)


def validateVPath(path):
    log.verbose(f"validating path='{path}'")
    (base, ext) = utils.splitPath(path)
    log.verbose(f"base={base}, ext={ext}")
    (vext) = getVPathExt(base + ext)
    log.verbose(f"vpath ext={vext}")
    if base == '':
        raise errors.VaultError(f"Invalid path '{path}' (base)")
    if ext == '':
        raise errors.VaultError(f"Invalid path '{path}' (ext)")
    if vext == '':
        raise errors.VaultError(f"Invalid path '{path}' (vext)")
    return (base + ext, base + ext + vext)


def validateOpenVPath(path):
    (plainPath, vpath) = validateVPaths(path)
    if os.path.exists(plainPath):
        raise errors.VaultError("'" + plainPath + "' exists")
    if not os.path.exists(vpath):
        raise errors.VaultError("'" + vpath + "' does not exist")
    return (plainPath, vpath)


def validateEditVPath(path):
    (plainPath, vpath) = validateVPaths(path)
    if os.path.exists(plainPath):
        raise errors.VaultError("'" + plainPath + "' exists")
    return (plainPath, vpath)


def validateEncryptVPath(path):
    (plainPath, vpath) = validateVPaths(path)
    if not os.path.exists(plainPath):
        raise errors.VaultError(f"'{plainPath}' does not exist")
    if os.path.exists(vpath):
        raise errors.VaultError(f"'{vpath}' exists")
    return (plainPath, vpath)

def validateReencryptVPath(path):
    (plainPath, vpath) = validateVPaths(path)
    if not os.path.exists(vpath):
        raise errors.VaultError(f"'{vpath}' does not exist")
    if os.path.exists(plainPath):
        raise errors.VaultError(f"'{plainPath}' exists")
    return (plainPath, vpath)


def getVPathExt(path):
    for vext in config.CONFIG['internal']['vexts']:
        vpath = path + vext
        log.verbose(f"checking for vpath {vpath}")
        if os.path.exists(vpath):
            log.verbose(f"found {vpath}")
            return vext
    log.verbose("no files found; using default extension")
    return config.CONFIG['internal']['vext.default']


def openVPath(path, destpath, cmd):
    log.verbose(f"openVPath {path} {destpath} {cmd}")
    (plainPath, vpath) = validateOpenVPath(path)
    pp64 = getPassPhrase(config.CONFIG['general']['group'], False)
    crypto.decryptFile(vpath, destpath, pp64)
    if destpath is not None and cmd is not None:
        if utils.runCommand(cmd, destpath):
            log.warning("file has been modified; updates discarded")


def editVPath(path, destpath, cmd):
    log.verbose(f"editVPath {path} {destpath} {cmd}")
    (plainPath, vpath) = validateEditVPath(path)
    log.verbose(f"editVpath plainPath={plainPath}, vpath={vpath}")
    if os.path.exists(vpath):
        log.verbose(f"editVpath vpath {vpath} exists; decrypting")
        pp64 = getPassPhrase(config.CONFIG['general']['group'], False)
        crypto.decryptFile(vpath, destpath, pp64)
    else:
        log.verbose(f"editVpath vpath {vpath} does not exist")
        pp64 = getPassPhrase(config.CONFIG['general']['group'], True)
    if utils.runCommand(cmd, destpath):
        file.backupFile(path)
        crypto.encryptFile(destpath, vpath, pp64)
    else:
        if os.path.exists(destpath):
            log.warning("file has not been modified")
        else:
            log.warning("file has not been created")
    file.deletePath(destpath)


def encryptVPath(path):
    log.verbose(f"encryptVPath path={path}")
    (plainPath, vpath) = validateEncryptVPath(path)
    log.verbose(f"encryptVpath plainPath={plainPath}, vpath={vpath}")
    pp64 = getPassPhrase(config.CONFIG['general']['group'], True)
    crypto.encryptFile(plainPath, vpath, pp64)
    file.deletePath(plainPath)



def reencryptVPath(path, destpath):
    log.verbose(f"reencryptVPath path={path}, destpath={destpath}")
    (plainPath, vpath) = validateReencryptVPath(path)
    log.verbose(f"reencryptVpath plainPath={plainPath}, vpath={vpath}")
    group = config.CONFIG['general']['group']
    group_current = group + ".current"
    group_new = group + ".new"

    pp64 = getPassPhrase(group_current, False, " (current)")
    crypto.decryptFile(vpath, destpath, pp64)
    
    pp64 = getPassPhrase(group_new, True, " (new)")
    crypto.encryptFile(destpath, vpath, pp64)


def getPassPhrase(group, confirm, tag = ""):
    log.verbose(f"getting passphrase group={group}, comfirm={confirm}")
    pp = client.sendRequest(['get', group])
    if len(pp) >= 1:
        log.verbose("got passphrase from server")
        log.sensitive(f"passphrase (base64): {pp[0]}")
        return pp[0]
    pp = ''
    try:
        while pp == '':
            pp = getpass.getpass(f"Passphrase{tag}: ")
        if confirm:
            pp2 = ''
            while pp2 == '':
                pp2 = getpass.getpass(f"Confirm Passphrase{tag}: ")
            if pp != pp2:
                del pp
                del pp2
                raise errors.VaultSecurityError("Passphrases do not match")
            del pp2
        log.sensitive(f"passphrase |{pp}|")
        pp64 = utils.str2b64(pp) # base64.b64encode(pp.encode('utf-8')).decode('utf-8')
        log.sensitive(f"base64 |{pp64}|")
        client.sendRequest(['set', group, pp64])
        del pp
        return pp64
    except EOFError as e:
        log.verbose(f"exception raised: {e}")
        raise errors.VaultQuit(str(e))
    except Exception as e:
        log.verbose(f"exception raised: {e}")
        raise errors.VaultQuit(str(e))



#
# MIT License
#
# Copyright (c) 2017-2019 Paul Taylor
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import socket
import re
import os
import subprocess
import time
import sys
import socket
import errno


from gpg_vault import config, vault, log, errors, utils

def killServer():
    log.info("killing server")
    sendRequest(['quit'])

def connect(sock, port):

    try:
        log.verbose(f"connecting to server on port {port}")
        sock.connect(('localhost', int(port)))
    except socket.error as e:
        if e.errno == errno.ECONNREFUSED:
            return False
        log.error(f"error connecting to server: {e}")
        raise e
    return True   

def sendRequest(args):

    log.verbose(f"sendRequest args={args}")
    req = " ".join(args)
    tryAgain = True
    while tryAgain:
        tryAgain = False
        port = getServerPort(utils.getVaultDir())
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            if connect(sock, port):
                log.verbose("sending request to server")
                sock.sendall((req + "\n").encode('utf-8'))
                reply = sock.recv(1024) 
            else:
                # Assume the server is not running
                log.info("cannot connect; trying again")
                time.sleep(1)
                deleteServerPort(utils.getVaultDir())
                tryAgain = True;
                continue
        finally:
            sock.close()

    ret = re.split('\s+', reply.decode('utf-8'))
    if ret is None or len(ret) < 1:
        raise VaultError('unable to process reply')

    log.sensitive(f"response: {ret}")
    if ret[0] == 'ok':
        log.verbose('reply ok')
        return ret[1:]
    elif ret[0] == 'error':
        log.verbose(f"reply {ret}")
        raise errors.VaultSecurityError(str(ret))
    else:
        log.verbose(f"unexpected reply {ret}")
        raise errors.VaultError('unable to process reply')


def getServerPort(vdir):

    log.verbose(f"getServerPort vdir={vdir}")
    pidFile = os.path.join(vdir, 'server.pid')
    if not os.path.exists(pidFile):
        log.verbose(f"pidfile {pidFile} does not exist")
        log.info("launching server")
        args = ['gpg-vault-server', vdir]
        global pid
        pid = subprocess.Popen(args=args, shell=False).pid
        log.verbose(f"pid: {pid}")

    count = 10
    while not os.path.exists(pidFile):
        log.info("waiting for server to start")
        time.sleep(1)

    f = open(pidFile, "r")
    lines = f.readlines()
    if lines is None or len(lines) != 2:
        log.error(f"cannot read pidFile {pidFile}")
        exit(1)

    port = int(lines[1])
    log.verbose(f"found server port={port}")
    return port

def deleteServerPort(vdir):
    pidFile = os.path.join(vdir, 'server.pid')
    os.unlink(pidFile)



#
# MIT License
#
# Copyright (c) 2017-2019 Paul Taylor
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import socketserver
import datetime
import re
import sys
import os
import time
import signal

from gpg_vault import config, utils, __version__

HOST = "localhost"
PMAP = {}

global logFile


def start():


    vdir = utils.getVaultDir()
    if not os.path.exists(vdir):
       print(f"ERROR - directory {vdir} does not exist")
       exit(1)

    signal.signal(signal.SIGTERM, shutdown_handler)
    signal.signal(signal.SIGINT, shutdown_handler)
    signal.signal(signal.SIGHUP, shutdown_handler)
    signal.signal(signal.SIGQUIT, shutdown_handler)

    logFilePath = os.path.join(vdir, 'server.log')
    global logFile
    logFile = open(logFilePath, "a")
    log_to_file("")
    log_to_file(f"starting; pid {os.getpid()}")
    log_to_file(f"version: {__version__}")

    pidFile = os.path.join(vdir, 'server.pid')
    if os.path.exists(pidFile):
        print(f"ERROR - file {pidFile} exists")
        exit(1)

    port = int(config.CONFIG['server']['port'])
    log_to_file(f"port {port}")

    timeout_reset_all = int(config.CONFIG['server']['timeout_reset_all'])
    log_to_file(f"timeout_reset_all {timeout_reset_all}")

    f = open(pidFile, "w")
    f.write(str(os.getpid()) + "\n")
    f.write(str(port) + "\n")
    f.close()


    server = VServer((HOST, port), timeout_reset_all, MyTCPHandler)

    global notShutdown
    notShutdown = True
    while notShutdown:
        log_to_file("waiting for request")
        server.handle_request()

    shutdown("requested")

def shutdown(msg):
    vdir = utils.getVaultDir()
    log_to_file(f"shutting down: {msg}")
    
    pidFile = os.path.join(vdir, 'server.pid')
    if os.path.exists(pidFile):
        log_to_file(f"removing PID file {pidFile}")
        os.unlink(pidFile)
    log_to_file("exiting")
    sys.exit(0)
     

def shutdown_handler(sig, frame):
    log_to_file(f"signal {sig}")
    shutdown(f"signal {sig}")

def reset():
    global PMAP
    log_to_file("reset")
    PMAP.clear()
    return "ok"


def request_set_passphrase(group, pw):
    global PMAP
    ts = time.time()
    log_to_file(f"request_set_passphrase group={group}, timestamp={ts}")

    PMAP[group] = (pw, ts)
    return "ok"


def request_get_passphrase(group):
    global PMAP
    ts = time.time()
    log_to_file(f"request_get_passphrase group={group}, timestamp={ts}")
    if group in PMAP:
        log_to_file("have cached passphrase")
        p = PMAP[group]
        tdiff = ts - p[1]
        log_to_file(f"time in use: {tdiff}")
        del PMAP[group]
        max_time = int(config.CONFIG['server']['timeout_reset_one'])
        if tdiff > max_time:
            log_to_file(f"timeout {tdiff} > {max_time}; removing from cache")
            return "ok"
        PMAP[group] = (p[0], ts)
        return "ok " + p[0]
    else:
        log_to_file("no password cached")
        return "ok"


def request_error(msg):
    log_to_file("ERROR: " + msg)
    reset()
    return "error " + msg

def request_ping():
    log_to_file("ping")
    return "ok"


def request_quit():
    log_to_file("quit")
    reset()
    global notShutdown
    notShutdown = False
    return "ok"


def request_reset():
    reset()
    return "ok"


def log_to_file(m):
    now = datetime.datetime.now()
    global logFile
    logFile.write(f"[{now}] {m}\n")
    logFile.flush()


class MyTCPHandler(socketserver.BaseRequestHandler):

    def __init__(self, request, client_address, server):
        log_to_file(f"request from client {client_address}")
        socketserver.BaseRequestHandler.__init__(self, request, client_address, server)

    def setup(self):
        log_to_file("setup")
        return socketserver.BaseRequestHandler.setup(self)

    def handle(self):
        request = self.request.recv(1024).strip()
        if self.client_address[0] != "127.0.0.1":
            reply = request_error("client not local")
        else:
            reply = self.processCmd(request)
            self.request.sendall(reply.encode('utf-8'))

    def processCmd(self, request):
        cmds = re.split('\s+', request.decode('utf-8'))
        if len(cmds) < 1:
            return request_error("invalid command")
        if cmds[0] == "set":
            if len(cmds) != 3:
                return request_error("invalid command")
            return request_set_passphrase(cmds[1], cmds[2])
        elif cmds[0] == "get":
            if len(cmds) != 2:
                return request_error("invalid command")
            return request_get_passphrase(cmds[1])
        elif cmds[0] == "reset":
            return request_reset()
        elif cmds[0] == "quit":
            return request_quit()
        elif cmds[0] == "ping":
            return request_ping()
        else:
            return request_error("unknown command")


class VServer(socketserver.TCPServer):

    allow_reuse_address = True
    timeout = 60 * 10  # TODO get this from config

    def __init__(self, server_addr, t, handlerClass):
        timeout = t  # !! NOT WORKING
        socketserver.TCPServer.__init__(self, server_addr, handlerClass)

    def handle_timeout(self):
        log_to_file("timeout")
        reset()



#
# MIT License
#
# Copyright (c) 2017-2019 Paul Taylor
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os
import sys
import traceback
import subprocess
import base64

from gpg_vault import log, config, utils

def isTrue(b):
    if isinstance(b, bool):
        return b
    else:
        return b.lower() in ('true', 'yes')


def isFalse(b):
    return not isTrue(b)


def str2b64(s):
    return base64.b64encode(s.encode('utf-8')).decode('utf-8')

def b642str(s):
    return base64.b64decode(s).decode('utf-8')
   
def runCommand(cmdName, arg):
    log.verbose(f"runCommand {cmdName} {arg}")
    before_mtime = 0

    cmd = config.CONFIG['commands'][cmdName]
    if cmd is None:
        log.error("no command configured for 'deleteFile'");
        log.warning(f"path {path} was not deleted")
        return

    cmdArgs = cmd.replace("{}", arg).split()
    log.verbose(f"cmdArgs: {cmdArgs}")
    if os.path.exists(arg):
        before_mtime = os.path.getmtime(arg)
    try:
        ret = subprocess.call(args=cmdArgs, shell=False)
        log.verbose(f"ret: {ret}")
        if ret != 0:
            log.warning("process exited with error")
    except Exception as e:
        log.error(f"error running command '{cmd} {arg}' -> {e}")
        return False

    if not os.path.exists(arg):
        return False
    after_mtime = os.path.getmtime(arg)
    log.verbose(f"before mtime: {before_mtime}")
    log.verbose(f"after mtime: {after_mtime}")
    return after_mtime > before_mtime


def getVaultDir():

    home = os.path.expanduser("~")

    log.verbose(f"home dir: {home}")

    if not os.path.exists(home):
        error(f"Directory {home} does not exist")

    t = config.WORK_DIR_NAME
    vdir = os.path.join(home, t)
    log.verbose(f"vault directory: {vdir}")

    if not os.path.exists(vdir):
        log.verbose(f"creating directory {vdir}")
        os.mkdir(vdir)
        if not os.path.exists(vdir):
            log.error(f"failed to create directory {vdir}")

    return vdir


def getTmpDir():

    td = getVaultDir()
    prefix = config.CONFIG['internal']['tmp.dir.prefix']
    p = "%s%s" % (prefix, str(os.getpid()))
    path = os.path.join(td, p)
    log.verbose(f"tmp tmp dir: {path}")
    if os.path.exists(path):
        log.error(f"tempory directory {path} already exists")
    os.mkdir(path)
    if not os.path.exists(path):
        log.error(f"tempory directory {path} was not created")
    return path


def oswarning():
    # TODO - get message from config
    if sys.platform == 'darwin':
        log.info("Use CMD-K to clear sensitve information from the screen")
    else:
        log.ingo("Use ^L to clear sensitve information from the screen")


def splitPath(path):
    (base, ext) = os.path.splitext(path)
    if ext == '':
        ext = config.CONFIG['internal']['ext.default']
    if ext in config.CONFIG['internal']['vexts']:
        (base, ext) = os.path.splitext(base)
    return (base, ext)


def log_exception(e):
    if utils.isTrue(config.CONFIG['general']['verbose']):
      tb = traceback.format_exc()
      log.error(tb)


#
# Merge two dicts recursively
#
def merge(a, b):
    c = {}
    for k in a:
        c[k] = a[k]

    for k in b:
        if k in c:
            v1 = c[k]
            v2 = b[k]
            if isinstance(v1, dict) and isinstance(v2, dict):
                c[k] = merge(v1, v2)
            else:
                c[k] = v2
        else:
            c[k] = b[k]

    return c


#
# dict2str
#
def dict2str(d):
    return dict2str2(d, 4)


def dict2str2(d, i):
    s = ""
    for k in d:
        s = s + "%s%s: " % (' ' * i, k)
        v = d[k]
        if isinstance(v, dict):
            s = s + '{\n'
            s = s + dict2str2(v, i + 4)
            s = s + "%s}\n" % (' ' * i)
        else:
            s = s + str(v)
            s = s + '\n'
    return s



#!/usr/bin/env python3
#
# MIT License
#
# Copyright (c) 2017-2019 Paul Taylor
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os
import ast

import setuptools
#from distutils.core import setup

def get_version():
    file_path = os.path.join(os.path.dirname(__file__), 'gpg_vault', '__init__.py')
    with open(file_path) as file_obj:
        root_node = ast.parse(file_obj.read())
        for node in ast.walk(root_node):
            if isinstance(node, ast.Assign):
                if len(node.targets) == 1 and node.targets[0].id == "__version__":
                    return node.value.s
    raise RuntimeError("Unable to find version string.")

setuptools.setup(
        name='gpg-vault',
        version=get_version(),
        description='Simple GPG based file encryption utility.',
        author='Paul Taylor',
        author_email='pftylr@gmail.com',
        url='https://github.com/ptaylor/gpg-vault',
        packages=[
                'gpg_vault'
        ],
        scripts=[
                'scripts/vcat',
                'scripts/vedit',
                'scripts/vopen',
                'scripts/vencrypt',
                'scripts/vreencrypt',
                'scripts/vclear',
                'scripts/gpg-vault-server'
        ],
        install_requires=[
        ],
        python_requires = '>3.7'
)



#
# MIT License
#
# Copyright (c) 2017-2019 Paul Taylor
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os

from gpg_vault import log, config, vault, utils, errors, client, file

def run(cmd, files):

    client.sendRequest(['ping'])

    if files is None:
        runForFile(cmd, None)
    else:
        for file in config.CONFIG['files']:
            log.verbose("running command '" + cmd + "' on file '" + file + "'")
            runForFile(cmd, file)

    if utils.isTrue(config.CONFIG['general']['kill_server']):
        client.killServer()

    exit(0)


def runForFile(cmd, file):

    try:
        {
            'cat': cmd_cat,
            'open': cmd_open,
            'edit': cmd_edit,
            'encrypt': cmd_encrypt,
            'reencrypt': cmd_reencrypt,
            'clear': cmd_clear
            }[cmd](cmd, file)
    except errors.VaultSecurityError as e:
        cmd_reset()
        log.error(str(e))
        exit(1)
    except errors.VaultError as e:
        log.error(str(e))
        exit(1)
    except errors.VaultQuit as e:
        utils.log_exception(e)
        log.error(str(e))
        exit(1)
    except Exception as e:
        utils.log_exception(e)
        log.error(f"Unexpected error. {e}")
        exit(1)


def cmd_cat(cmd, path):
    log.verbose(f"cmd_cat {cmd}, {path}")
    vault.openVPath(path, None, None)
    print("")
    utils.oswarning()


def cmd_open(cmd, path):
    tmpdir = utils.getTmpDir()
    try:
        tmppath = os.path.join(tmpdir, str(os.getpid()))
        log.verbose(f"cmd_open {cmd}, {path}")
        vault.openVPath(path, tmppath, cmd)
    finally:
        file.deleteDir(tmpdir)


def cmd_edit(cmd, path):
    log.verbose(f"cmd_edit {cmd}, {path}")
    tmpdir = utils.getTmpDir()
    try:
        tmppath = os.path.join(tmpdir, str(os.getpid()))
        vault.editVPath(path, tmppath, cmd)
    finally:
        file.deleteDir(tmpdir)


def cmd_encrypt(cmd, path):
    log.verbose(f"cmd_encrypt cmd={cmd}, path={path}")
    vault.encryptVPath(path)

def cmd_reencrypt(cmd, path):
    log.verbose(f"cmd_reencrypt cmd={cmd}, path={path}")
    tmpdir = utils.getTmpDir()
    try:
        tmppath = os.path.join(tmpdir, str(os.getpid()))
        vault.reencryptVPath(path, tmppath)
    finally:
        file.deleteDir(tmpdir)


def cmd_clear(cmd, path):
    log.verbose(f"cmd_clear {cmd}, {path}")
    cmd_reset()


def cmd_unknown(cmd, path):
    usage(f"unknown option '{cmd}'")


def cmd_reset():
    log.verbose("sending reset request")
    client.sendRequest(['reset'])
